@@ -1,16 +1,15 @@
+import builtins
+import functools
+import glob
 import importlib.util
 import inspect
 import logging
 import os
-import random
 import warnings
 from contextlib import contextmanager
 from socket import gethostname
-from string import ascii_letters
 from typing import Optional, Type, TypeVar
 
-import functools
-import builtins
 import yaml
 from pydantic import ValidationError
 from pydantic.main import BaseModel
@@ -84,11 +83,28 @@ def debug():
     else:
         return version() == "unknown" and not _inside_docker()
 
-def bridged_slave():
-    if "BRIDGED_SLAVE" in os.environ:
-        return bool("BRIDGED_SLAVE", default=False)
+
+def compute_id():
+    return int("COMPUTE_ID", 1)
+
+
+def dist_folder():
+    if testing():
+        return string("DIST_FOLDER", "../tests/dist2")
+    elif debug():
+        return string("DIST_FOLDER", "../dist2")
     else:
-        return None
+        return string("DIST_FOLDER", "/opt/openmodule/dist2")
+
+
+def dev_device():
+    if "DEVICE_HOST" in os.environ:
+        return "test" in string("DEVICE_HOST", "")
+    elif testing() or debug():
+        return True
+    else:
+        return False
+
 
 def settings_module():
     if testing() and os.path.exists("../tests/config.py"):
@@ -199,6 +215,7 @@ def config_yaml_path():
 
 def yaml(model: Type[YamlType], path: str = None) -> YamlType:
     yaml_path = path or config_yaml_path()
+
     try:
         if os.path.exists(yaml_path):
             import yaml
@@ -408,6 +425,16 @@ class OverrideSettings:
         return overrides
 
 
+def locale_dir():
+    if "LOCALE_DIR" in os.environ:
+        return string("DEBUG")
+    elif testing() or debug():
+        # for libraries the translations are in package folder (e.g. in ../libcontroller/translation/locale)
+        return next(iter(glob.glob("../*/translation/locale")), "../docker/translation/locale")
+    else:
+        return "/translation/locale"
+
+
 @contextmanager
 def override_context(**options):
     settings.override(**options)
@@ -440,6 +467,7 @@ def override_settings(**options):
 class GlobalSettings:
     # usual
     NAME = string("NAME", "om_dev_unnamed_1")
+    PARENT = string("PARENT", "") or None
     VERSION = version()
     RESOURCE = resource()
     DEBUG = debug()
@@ -447,13 +475,38 @@ class GlobalSettings:
     LOG_LEVEL = log_level()
     DATABASE_FOLDER = database_folder()
 
+    CURRENCY = string("CURRENCY", "EUR")
+    TIMEZONE = string("TIMEZONE", "Europe/Vienna")
+
     # broker env vars
     BROKER_SUB = broker_sub()
     BROKER_PUB = broker_pub()
 
     LOCAL_DEVELOPMENT = bool("LOCAL_DEVELOPMENT", False)
-    BRIDGED_SLAVE = bridged_slave()
+    COMPUTE_ID = compute_id()
+    DAEMON = bool("DAEMON", False)
+    DIST_FOLDER = dist_folder()
+    DEV_DEVICE = dev_device()
+
+    # redis
+    REDIS_HOST = string("REDIS_HOST", "localhost")
+    REDIS_PASSWORD = string("REDIS_PASSWORD", "") or None
+    REDIS_PORT = int("REDIS_PORT", 6379)
+    REDIS_DB = int("REDIS_DB", 0)
+    REDIS_CONNECTION_STATUS_KEY = string("REDIS_CONNECTION_STATUS_KEY", "device_connection_status")
+
+    # translation
+    LOCALE_DIR = locale_dir()
+    LANGUAGE = "" if testing() else string("LANGUAGE", "de").lower()
 
 
 settings = LazySettings()
 _override_settings = OverrideSettings()
+
+
+def is_bridged_master():
+    return settings.COMPUTE_ID == 1
+
+
+def is_bridged_slave():
+    return not is_bridged_master()
