@@ -1,15 +1,17 @@
 import base64
 import binascii
 import re
+import warnings
 from datetime import datetime
 from decimal import Decimal
 from enum import Enum
 from json.encoder import ESCAPE_ASCII
+from typing import Optional, Union
 
 import orjson
 import zmq
 from dateutil.tz import UTC
-from pydantic import Field, BaseModel, validator
+from pydantic import Field, BaseModel, validator, root_validator
 from pydantic.main import ROOT_KEY
 
 from openmodule import config
@@ -50,6 +52,10 @@ def _unicode_escape(data):
 def default(obj):
     if isinstance(obj, Decimal):
         return str(obj)
+    if settings.TESTING:
+        from freezegun.api import FakeDatetime
+        if isinstance(obj, FakeDatetime):
+            return obj.isoformat()
     raise TypeError
 
 
@@ -128,8 +134,14 @@ class ZMQMessage(OpenModuleModel):
 
     _tz_timestamp = timezone_validator("timestamp")
 
-    def publish_on_topic(self, pub_socket: zmq.Socket, topic: bytes):
-        pub_socket.send_multipart((topic, self.json_bytes()))
+    def publish_on_topic(self, pub_socket: zmq.Socket, topic: str):
+        # TODO: backwards compatibiltiy, remove in version 12
+        if isinstance(topic, bytes):
+            warnings.warn('\n\nPassing topic as bytes is deprecated. Please pass as strings.\n',
+                          DeprecationWarning, stacklevel=2)
+            topic = topic.decode("utf8")
+
+        pub_socket.send_multipart((topic.encode("utf8"), self.json_bytes()))
 
     def dict(self, **kwargs):
         data = super().dict(**kwargs)
@@ -149,3 +161,41 @@ class Gateway(OpenModuleModel):
 
     def __str__(self):
         return f"{self.gate}/{self.direction}"
+
+
+class GateType(str, Enum):
+    entry = "entry"
+    exit = "exit"
+    bidirectional = "bidirectional"
+    door = "door"
+    transit_barrier = "transit_barrier"
+    transit_freeflow = "transit_freeflow"
+    unknown = "unknown"
+
+    def is_outward(self):
+        return self.value in [GateType.exit, GateType.bidirectional]
+
+
+class Gate(OpenModuleModel):
+    gate: str
+    name: str = ""
+    type: GateType = GateType.unknown
+
+
+class CostEntryData(OpenModuleModel):
+    entry_type: str
+    value: Optional[Union[int, str]]
+    group: Optional[str]
+    account_id: str
+    source: Optional[str]  # if None, source will be set to service name on default: e.g. "service_iocontroller"
+    source_id: Optional[str]
+    idempotency_key: Optional[str]
+
+    @root_validator
+    def set_value_type(cls, values):
+        if values["entry_type"] in ["rate_change", "rate_validation", "partial_rate_change"]:
+            values["value"] = str(values["value"])
+        elif values["entry_type"] in ["static_cost", "pending_static_cost", "amount_validation", "time_validation"
+                                      "payment"]:
+            values["value"] = int(values["value"])
+        return values
