@@ -1,8 +1,9 @@
 import logging
+import os
 import threading
 import warnings
 from collections import namedtuple
-from typing import Callable, Dict, Optional, Type, Union, TypeVar
+from typing import Callable, Dict, Optional, Type, Union, TypeVar, Tuple
 
 import zmq
 from pydantic import ValidationError, BaseModel
@@ -11,7 +12,7 @@ from pydantic.main import ROOT_KEY
 from openmodule.config import settings
 from openmodule.dispatcher import DummyExecutor
 from openmodule.messaging import get_sub_socket, get_pub_socket, receive_message_from_socket
-from openmodule.models.rpc import RPCResponse, RPCRequest
+from openmodule.models.rpc import RPCResponse, RPCRequest, RPCServerError
 from openmodule.rpc.common import channel_to_request_topic, channel_to_response_topic
 from openmodule.threading import get_thread_wrapper
 from openmodule.utils.schema import Schema
@@ -59,7 +60,7 @@ class RPCServer(object):
         if config:
             warnings.warn(
                 "\n\nThe RPC Server no longer requires the config, but takes the openmodule global config instead. "
-                "Please remove the argument. Future versions will remove the argument and leat to an error",
+                "Please remove the argument. Future versions will remove the argument and lead to an error",
                 DeprecationWarning, stacklevel=2
             )
 
@@ -68,7 +69,7 @@ class RPCServer(object):
         self.sub = get_sub_socket(context, config)
         self.pub = get_pub_socket(context, config)
         self.pub_lock = threading.Lock()
-        self.handlers: Dict[tuple, HandlerEntry] = {}
+        self.handlers: Dict[Tuple[str, str], HandlerEntry] = {}
         self.filters = []
         self.running = True
         self.log = logging.getLogger(self.__class__.__name__)
@@ -89,7 +90,7 @@ class RPCServer(object):
 
     def add_filter(self, filter: Callable[[], bool], channel=None, type=None):
         """
-        :param filter: filter function with paramters reqeust, message, handler
+        :param filter: filter function with parameters request, message, handler
         :param channel: specific channel to filter or all channels if None
         :param type: specific type to filter or all types if None
         """
@@ -125,16 +126,16 @@ class RPCServer(object):
             raise ValueError(f"handler for {channel}:{type} is already registered")
 
         self.log.debug("register handler {}:{} -> {}".format(channel, type, handler))
-        self.sub.subscribe(channel_to_request_topic(channel.encode("ascii")))
+        self.sub.subscribe(channel_to_request_topic(channel).encode("utf8"))
         self.handlers[(channel, type)] = HandlerEntry(request_class, response_class, handler)
         self.registered_channels = set(x[0] for x in self.handlers.keys())
         if register_schema:
             Schema.save_rpc(channel, type, request_class, response_class, handler)
 
-    def _channel_from_topic(self, topic: bytes) -> str:
-        return topic.split(b"-", 2)[-1].decode("ascii")
+    def _channel_from_topic(self, topic: str) -> str:
+        return topic.split("-", 2)[-1]
 
-    def should_process_message(self, request, message, handler, channel):
+    def should_process_message(self, request, message, handler, channel: str):
         if self.filters:
             result = True
             for filter in self.filters:
@@ -160,33 +161,36 @@ class RPCServer(object):
         data = response.dict()
         if response.__custom_root_type__:
             data = data[ROOT_KEY]
+        if "status" in data:
+            self.log.error('"status" must not be a toplevel-key in RPCResponse. Terminating to alert the user!')
+            os._exit(112)
         return data
 
-    def process_rpc(self, channel, message: RPCRequest) -> Optional[Dict]:
+    def process_rpc(self, channel: str, message: RPCRequest) -> Optional[Dict]:
         if self.resource and message.resource and self.resource != message.resource:
-            self.log.debug("a rpc was not processed because the resource did not mach. "
+            self.log.debug("a rpc was not processed because the resource did not match. "
                            "my resource: %s, resource specified in rpc: %s", self.resource, message.resource)
             return None
 
         if settings.LOG_LEVEL == logging.DEBUG:
             self.log.debug("received RPC channel: %s, type: %s, request: %s", channel, message.type, message.request)
-        else:
-            self.log.info("received RPC channel: %s, type: %s", channel, message.type)
 
         handler: HandlerEntry = self.find_handler(channel, message.type)
         if handler:
             try:
                 request = handler.request_class(**message.request)
             except ValidationError as e:
-                return {"status": "validation_error", "exception": e.json()}
+                self.log.exception("exception in parsing request. request not processed and error returned")
+                return {"status": RPCServerError.validation_error, "exception": e.json()}
 
             try:
                 if not self.should_process_message(request, message, handler, channel):
                     return None
             except Exception as e:
                 self.log.exception("exception in message filter. request not processed and error returned")
-                return {"status": "filter_error", "exception": str(e)}
-
+                return {"status": RPCServerError.filter_error, "exception": str(e)}
+            if settings.LOG_LEVEL != logging.DEBUG:
+                self.log.info("received RPC channel: %s, type: %s", channel, message.type)
             try:
                 response = handler.handler(request, message)
                 if isinstance(response, handler.response_class):
@@ -198,7 +202,7 @@ class RPCServer(object):
                     return self._response_to_dict(response)
             except Exception as e:
                 self.log.exception("exception in handler {}:{}".format(channel, message.type))
-                return {"status": "handler_error", "exception": str(e)}
+                return {"status": RPCServerError.handler_error, "exception": str(e)}
         else:
             self.log.warning("no handler found for {}:{}".format(channel, message.type))
             return None
@@ -218,7 +222,7 @@ class RPCServer(object):
             except ValidationError as e:
                 message_type = "unknown"
                 rpc_id = None
-                response = {"status": "validation_error", "exception": e.json()}
+                response = {"status": RPCServerError.validation_error, "exception": e.json()}
             else:
                 response = self.process_rpc(channel, message)
 
@@ -236,14 +240,14 @@ class RPCServer(object):
                 with self.pub_lock:
                     result.publish_on_topic(
                         self.pub,
-                        channel_to_response_topic(channel.encode("ascii")),
+                        channel_to_response_topic(channel),
                     )
-        except Exception:
-            self.log.exception("Unknown exception while processing an rpc. No answer could was sent to the caller.")
         except zmq.ContextTerminated:
             self.log.warning(
                 "context terminated while RPC request was being processed. The result might not have been sent."
             )
+        except Exception:
+            self.log.exception("Unknown exception while processing an rpc. No answer could was sent to the caller.")
 
     def run(self):
         poller = zmq.Poller()
