@@ -25,6 +25,7 @@ class ExceptionProcess(Process):
     def run(self):
         try:
             super().run()
+            Schema.to_file()
             self.is_finished.set()
         except Exception as e:
             logging.exception(e)
@@ -131,7 +132,7 @@ class InterruptTestMixin(TestCase):
                 raise AssertionError("Process did not finish gracefully")
 
     def send_signal_to_process(self, process: ExceptionProcess, signal: Union[Type[KeyboardInterrupt], int]):
-        self.assertFalse(process.is_finished.is_set())
+        self.assertFalse(process.is_finished.is_set(), msg="process crashed before we could send the signal")
         if signal == KeyboardInterrupt:
             signal = SIGINT
         os.kill(process.pid, signal)
@@ -142,16 +143,24 @@ class InterruptTestMixin(TestCase):
         wraps the helper functions into a single function because apparently everybody likes
         to write framework functions which are hard to adopt and do one single thing and nothing
         else. mostly here for backwards compatibility and reference how to use the testcase
+        :return: list of output lines
         """
-        process = self.start_process(f)
-        self.wait_for_setup()
-        time.sleep(raise_exception_after)
-        self.send_signal_to_process(process, signal)
-        self.assertCleanShutdown(process, shutdown_timeout)
+        with self.assertLogs() as cm:
+            try:
+                process = self.start_process(f)
+                self.wait_for_setup()
+                time.sleep(raise_exception_after)
+                self.send_signal_to_process(process, signal)
+                self.assertCleanShutdown(process, shutdown_timeout)
+            except Exception as e:
+                for line in cm.output:
+                    print(line)
+                raise e
+        return cm.output
 
 
 class MainTestMixin(HealthTestMixin, InterruptTestMixin):
-    topics = ["healthz"]
+    topics = ["healthpong"]
     protocol = "tcp://"
 
     def wait_for_setup(self):
