@@ -1,14 +1,17 @@
 import logging
+import os
+import threading
 import time
 from concurrent.futures import ThreadPoolExecutor
 from functools import partial
 from typing import Union, Callable
 from unittest import TestCase
+from unittest.mock import MagicMock
 from uuid import uuid4
 
 from pydantic.types import conint
 
-from openmodule.config import settings, override_context
+from openmodule.config import settings
 from openmodule.models.base import OpenModuleModel, EmptyModel, ZMQMessage
 from openmodule.rpc import server
 from openmodule.rpc.client import RPCClient
@@ -34,8 +37,8 @@ def _fake_request(gate=..., direction=...):
 
 class RPCCommonTestCase(TestCase):
     def test_topics(self):
-        self.assertEqual(b"rpc-rep-test", channel_to_response_topic(b"test"))
-        self.assertEqual(b"rpc-req-test", channel_to_request_topic(b"test"))
+        self.assertEqual("rpc-rep-test", channel_to_response_topic("test"))
+        self.assertEqual("rpc-req-test", channel_to_request_topic("test"))
 
     def test_gateway_filter(self):
         filter_full = server.gateway_filter("gate1", "in")
@@ -77,6 +80,11 @@ class TestRPCRequest(OpenModuleModel):
     pass
 
 
+class TestRPCRequest2(OpenModuleModel):
+    __test__ = False
+    some_data: str
+
+
 class TestRPCResponse(OpenModuleModel):
     __test__ = False
     pass
@@ -87,6 +95,10 @@ class TestRPCResponse2(OpenModuleModel):
     some_payload: str
 
 
+class RPCResponseWithStatus(OpenModuleModel):
+    status = "bad"
+
+
 class RPCServerTestCase(RPCServerTestMixin):
     rpc_channels = ["channel", "channel2", "a", "ab"]
     topics = ["test"]
@@ -118,10 +130,10 @@ class RPCServerTestCase(RPCServerTestMixin):
         rpc_request = RPCRequest(name="testclient", type="test-type", rpc_id=uuid4(), request={})
         broken_data = rpc_request.dict()
         del broken_data["rpc_id"]
-        self.zmq_client.send(f"rpc-req-channel".encode("ascii"), broken_data)
+        self.zmq_client.send(f"rpc-req-channel", broken_data)
 
         _, error_response = self.zmq_client.wait_for_message(
-            filter=lambda topic, message: topic == b"rpc-rep-channel" and message.get("type") == "unknown"
+            filter=lambda topic, message: topic == "rpc-rep-channel" and message.get("type") == "unknown"
         )
         self.assertIsNone(error_response.get("rpc_id"))
 
@@ -129,7 +141,7 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, self.set_called)
         self.wait_for_rpc_server(self.server)
 
-        self.rpc("channel", "test-type", {})
+        self.rpc("channel", "test-type", {}, TestRPCResponse)
         self.assertTrue(self.called_types.get("test-type"))
 
     def test_exception_in_filter_function(self):
@@ -140,9 +152,9 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.wait_for_rpc_server(self.server)
         self.server.add_filter(bad_filter)
 
-        response = self.rpc("channel", "test-type", {})
-        self.assertRPCFailure(response, "filter_error")
-        self.assertEqual("Error123", response["response"]["exception"])
+        with self.assertRaises(RPCClient.ServerFilterError) as e:
+            response = self.rpc("channel", "test-type", {}, TestRPCResponse)
+        self.assertEqual("Error123", str(e.exception))
 
     def test_filter(self):
         def allow_only_type_test(message: RPCRequest, **kwargs) -> bool:
@@ -152,13 +164,11 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.register_handler("channel", "test-type2", TestRPCRequest, TestRPCResponse, self.set_called)
         self.wait_for_rpc_server(self.server)
 
-        response = self.rpc("channel", "test-type", {})
-        self.assertRPCSuccess(response)
+        response = self.rpc("channel", "test-type", {}, TestRPCResponse)
         self.assertTrue(self.called_types.get("test-type"))
 
-        response = self.rpc("channel", "test-type2", {})
+        response = self.rpc("channel", "test-type2", {}, TestRPCResponse)
         self.assertTrue(self.called_types.get("test-type2"))
-        self.assertRPCSuccess(response)
 
         # clear stats, and add the filter
         del self.called_types["test-type"]
@@ -166,13 +176,12 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.add_filter(allow_only_type_test)
 
         # should still work as before
-        response = self.rpc("channel", "test-type", {})
-        self.assertRPCSuccess(response)
+        response = self.rpc("channel", "test-type", {}, TestRPCResponse)
         self.assertTrue(self.called_types.get("test-type"))
 
         # times out
-        with self.assertRaises(TimeoutError):
-            self.rpc("channel", "test-type2", {})
+        with self.assertRaises(RPCClient.TimeoutError):
+            self.rpc("channel", "test-type2", {}, TestRPCResponse, timeout=1)
         self.assertNotIn("test-type2", self.called_types)
 
     def test_filter_channel_type(self):
@@ -180,14 +189,13 @@ class RPCServerTestCase(RPCServerTestMixin):
             return False
 
         def check_ok():
-            response = self.rpc("channel", "test-type", {})
-            self.assertRPCSuccess(response)
+            response = self.rpc("channel", "test-type", {}, TestRPCResponse)
             self.assertTrue(self.called_types.get("test-type"))
             self.called_types.clear()
 
         def check_not_called():
-            with self.assertRaises(TimeoutError):
-                response = self.rpc("channel", "test-type", {})
+            with self.assertRaises(RPCClient.TimeoutError):
+                response = self.rpc("channel", "test-type", {}, TestRPCResponse, timeout=1)
             self.assertEqual(None, self.called_types.get("test-type"))
             self.called_types.clear()
 
@@ -222,9 +230,9 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, handler)
         self.wait_for_rpc_server(self.server)
 
-        response = self.rpc("channel", "test-type", {})
-        self.assertRPCFailure(response, "handler_error")
-        self.assertEqual("Error123", response["response"]["exception"])
+        with self.assertRaises(RPCClient.ServerHandlerError) as e:
+            response = self.rpc("channel", "test-type", {}, TestRPCResponse)
+        self.assertEqual("Error123", str(e.exception))
 
     def test_handler_returns_none(self):
         def handler(*_, **__):
@@ -234,8 +242,7 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, handler)
         self.wait_for_rpc_server(self.server)
 
-        response = self.rpc("channel", "test-type", {})
-        self.assertRPCSuccess(response)
+        response = self.rpc("channel", "test-type", {}, TestRPCResponse)
 
     def test_handler_returns_dict(self):
         def handler(*_, **__):
@@ -245,8 +252,7 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, handler)
         self.wait_for_rpc_server(self.server)
 
-        response = self.rpc("channel", "test-type", {})
-        self.assertRPCSuccess(response)
+        response = self.rpc("channel", "test-type", {}, TestRPCResponse)
 
     def test_handler_returns_model(self):
         def handler(*_, **__):
@@ -256,8 +262,7 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, handler)
         self.wait_for_rpc_server(self.server)
 
-        response = self.rpc("channel", "test-type", {})
-        self.assertRPCSuccess(response)
+        response = self.rpc("channel", "test-type", {}, TestRPCResponse)
 
     def test_handler_returns_wrong_dict(self):
         def bad_handler(*_, **__):
@@ -272,11 +277,10 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.register_handler("channel", "test-type2", TestRPCRequest, TestRPCResponse2, good_handler)
         self.wait_for_rpc_server(self.server)
 
-        response = self.rpc("channel", "test-type1", {})
-        self.assertRPCFailure(response, "handler_error")
+        with self.assertRaises(RPCClient.ServerHandlerError):
+            response = self.rpc("channel", "test-type1", {}, TestRPCResponse2)
 
-        response = self.rpc("channel", "test-type2", {})
-        self.assertRPCSuccess(response)
+        response = self.rpc("channel", "test-type2", {}, TestRPCResponse2)
 
     def test_handler_returns_wrong_response_type(self):
         def bad_handler(*_, **__):
@@ -291,11 +295,10 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.register_handler("channel", "test-type2", TestRPCRequest, TestRPCResponse2, good_handler)
         self.wait_for_rpc_server(self.server)
 
-        response = self.rpc("channel", "test-type1", {})
-        self.assertRPCFailure(response, "handler_error")
+        with self.assertRaises(RPCClient.ServerHandlerError):
+            response = self.rpc("channel", "test-type1", {}, TestRPCResponse2)
 
-        response = self.rpc("channel", "test-type2", {})
-        self.assertRPCSuccess(response)
+        response = self.rpc("channel", "test-type2", {}, TestRPCResponse2)
 
     def test_handler_returns_additional_data(self):
         """
@@ -309,9 +312,8 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.register_handler("channel", "test-type1", TestRPCRequest, TestRPCResponse2, too_much_handler)
         self.wait_for_rpc_server(self.server)
 
-        response = self.rpc("channel", "test-type1", {})
-        self.assertRPCSuccess(response)
-        self.assertDictEqual(response["response"], {"some_payload": "test", "status": "ok"})
+        response = self.rpc("channel", "test-type1", {}, TestRPCResponse2)
+        self.assertEqual(response, TestRPCResponse2(some_payload="test", status="ok"))
 
     def test_serializer_error(self):
         class SomeValidator(OpenModuleModel):
@@ -324,10 +326,12 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.register_handler("channel", "test-type", SomeValidator, TestRPCResponse, handler)
         self.wait_for_rpc_server(self.server)
 
-        self.assertRPCSuccess(self.rpc("channel", "test-type", {"max_int": 1}))
-        self.assertRPCSuccess(self.rpc("channel", "test-type", {"max_int": 5}))
-        self.assertRPCFailure(self.rpc("channel", "test-type", {"max_int": 0}), "validation_error")
-        self.assertRPCFailure(self.rpc("channel", "test-type", {"max_int": 10}), "validation_error")
+        self.rpc("channel", "test-type", {"max_int": 1}, TestRPCResponse)
+        self.rpc("channel", "test-type", {"max_int": 5}, TestRPCResponse)
+        with self.assertRaises(RPCClient.ServerValidationError):
+            self.rpc("channel", "test-type", {"max_int": 0}, TestRPCResponse)
+        with self.assertRaises(RPCClient.ServerValidationError):
+            self.rpc("channel", "test-type", {"max_int": 10}, TestRPCResponse)
 
     def test_no_handler_found(self):
         def handler(*_, **__):
@@ -337,17 +341,16 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.server.register_handler("channel", "test-type", TestRPCRequest, TestRPCResponse, handler)
         self.wait_for_rpc_server(self.server)
 
-        response = self.rpc("channel", "test-type", {})  # correct type and channel
-        self.assertRPCSuccess(response)
+        response = self.rpc("channel", "test-type", {}, TestRPCResponse)  # correct type and channel
 
-        with self.assertRaises(TimeoutError):  # wrong type
-            self.rpc("channel", "test-type-doesnotexist", {})
+        with self.assertRaises(RPCClient.TimeoutError):  # wrong type
+            self.rpc("channel", "test-type-doesnotexist", {}, TestRPCResponse, timeout=1)
 
-        with self.assertRaises(TimeoutError):  # wrong channel (correct prefix chosen on purpose)
-            self.rpc("channel2", "test-type", {})
+        with self.assertRaises(RPCClient.TimeoutError):  # wrong channel (correct prefix chosen on purpose)
+            self.rpc("channel2", "test-type", {}, TestRPCResponse, timeout=1)
 
-        with self.assertRaises(TimeoutError):  # wrong everything
-            self.rpc("channel2", "test-type-doesnotexist", {})
+        with self.assertRaises(RPCClient.TimeoutError):  # wrong everything
+            self.rpc("channel2", "test-type-doesnotexist", {}, TestRPCResponse, timeout=1)
 
     def test_correct_resource(self):
         def handler1(*_, **__):
@@ -359,17 +362,15 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.wait_for_rpc_server(self.server)
 
         with self.assertLogs() as cm:
-            response = self.rpc("channel", "test-type", {}, resource=None)
-        self.assertRPCSuccess(response)
+            response = self.rpc("channel", "test-type", {}, TestRPCResponse, resource=None)
         self.assertIn("handler 1", str(cm.output))
 
         with self.assertLogs() as cm:
-            response = self.rpc("channel", "test-type", {}, resource=settings.RESOURCE)
-        self.assertRPCSuccess(response)
+            response = self.rpc("channel", "test-type", {}, TestRPCResponse, resource=settings.RESOURCE)
         self.assertIn("handler 1", str(cm.output))
 
-        with self.assertRaises(TimeoutError):
-            self.rpc("channel", "test-type", {}, resource="test-2")
+        with self.assertRaises(RPCClient.TimeoutError):
+            self.rpc("channel", "test-type", {}, TestRPCResponse, resource="test-2", timeout=1)
 
     def test_prefix_matched_topic(self):
         def handler1(*_, **__):
@@ -385,8 +386,8 @@ class RPCServerTestCase(RPCServerTestMixin):
             # so we log some dummy message and check assertNotIn afterwards
             logging.info("some-log-so assert works")
 
-            with self.assertRaises(TimeoutError):
-                self.rpc("ab", "test-type", {})
+            with self.assertRaises(RPCClient.TimeoutError):
+                self.rpc("ab", "test-type", {}, TestRPCResponse, timeout=1)
 
         self.assertNotIn("no handler found", str(cm.output))
 
@@ -432,27 +433,24 @@ class RPCServerTestCase(RPCServerTestMixin):
         self.wait_for_rpc_server(self.server)
 
         # dict, should work
-        response = self.rpc("channel", "dict", {}, resource=None)
-        self.assertRPCSuccess(response)
-        self.assertEqual(response["response"], {"some_str": "str", "status": "ok"})
+        response = self.rpc("channel", "dict", {}, response_type, resource=None)
+        self.assertEqual(response, response_type(__root__=String(some_str="str")))
 
         # string response, should work
-        response = self.rpc("channel", "string", {}, resource=None)
-        self.assertRPCSuccess(response)
-        self.assertEqual(response["response"], {"some_str": "1", "status": "ok"})
+        response = self.rpc("channel", "string", {}, response_type, resource=None)
+        self.assertEqual(response, response_type(__root__=String(some_str="1")))
 
         # integer response, should work
-        response = self.rpc("channel", "integer", {}, resource=None)
-        self.assertRPCSuccess(response)
-        self.assertEqual(response["response"], {"some_int": 1, "status": "ok"})
+        response = self.rpc("channel", "integer", {}, response_type, resource=None)
+        self.assertEqual(response, response_type(__root__=Integer(some_int=1)))
 
         # none response, should not work, as required fields are not set
-        response = self.rpc("channel", "none", {}, resource=None)
-        self.assertGreater(response["response"].items(), {"status": "handler_error"}.items())
+        with self.assertRaises(RPCClient.ServerHandlerError):
+            response = self.rpc("channel", "none", {}, response_type, resource=None)
 
         # test_response, an incorrect response should not work
-        response = self.rpc("channel", "none", {}, resource=None)
-        self.assertGreater(response["response"].items(), {"status": "handler_error"}.items())
+        with self.assertRaises(RPCClient.ServerHandlerError):
+            response = self.rpc("channel", "none", {}, response_type, resource=None)
 
     def test_async_receive(self):
         """
@@ -464,24 +462,33 @@ class RPCServerTestCase(RPCServerTestMixin):
         def publishing_handler(*_, **__):
             """test handler which publishes a message"""
             message = ZMQMessage(type="abc")
-            message.publish_on_topic(self.server.pub, b"test")
+            message.publish_on_topic(self.server.pub, "test")
 
         self.server.register_handler("channel", "test", EmptyModel, EmptyModel, publishing_handler)
         self.wait_for_rpc_server(self.server)
 
         # this shows the current limitation in the test framework, if we ever do change the setup to better
         # support capturing messages from multiple topics, case will fail
-        response = self.rpc("channel", "test", EmptyModel())
-        self.assertRPCSuccess(response)
+        response = self.rpc("channel", "test", EmptyModel(), EmptyModel)
         with self.assertRaises(TimeoutError):
-            self.zmq_client.wait_for_message_on_topic(b"test")
+            self.zmq_client.wait_for_message_on_topic("test", timeout=1)
 
         # this shows how we can capture the message
-        rpc_id = self.rpc("channel", "test", EmptyModel(), receive_response=False)
-        message = self.zmq_client.wait_for_message_on_topic(b"test")
+        rpc_id = self.rpc_no_response("channel", "test", EmptyModel())
+        message = self.zmq_client.wait_for_message_on_topic("test")
         self.assertEqual(message["type"], "abc")
-        response = self.receive_rpc_response("channel", rpc_id)
-        self.assertRPCSuccess(response)
+        response = self.receive_rpc_response("channel", rpc_id, EmptyModel)
+
+    def test_status_in_response_model(self):
+        def handler(*_, **__):
+            """test handler which publishes a message"""
+            return RPCResponseWithStatus()
+        self.server.register_handler("channel", "test", EmptyModel, RPCResponseWithStatus, handler)
+        self.wait_for_rpc_server(self.server)
+        os._exit = MagicMock()
+        with self.assertRaises(RPCClient.ValidationError):
+            rpc_id = self.rpc("channel", "test", EmptyModel(), RPCResponseWithStatus)
+        assert os._exit.called
 
 
 class RpcClientTest(RPCServerTestMixin, OpenModuleCoreTestMixin):
@@ -500,70 +507,96 @@ class RpcClientTest(RPCServerTestMixin, OpenModuleCoreTestMixin):
         time.sleep(3)
         return TestRPCResponse2(some_payload="abc")
 
+    @staticmethod
+    def broken_filter(message: RPCRequest, **kwargs) -> bool:
+        assert False
+
+    @staticmethod
+    def broken_handler(*_, **__):
+        """test handler"""
+        assert False
+
     def setUp(self):
         super().setUp()
         self.server = RPCServer(context=self.zmq_context(), filter_resource=False)
         self.server.register_handler("channel", "test", TestRPCRequest, TestRPCResponse2, self.handler1)
+        self.server.register_handler("channel", "test2", TestRPCRequest2, TestRPCResponse2, self.handler1)
+        self.server.register_handler("channel", "test3", TestRPCRequest, TestRPCResponse2, self.handler1)
+        self.server.register_handler("channel", "test4", TestRPCRequest, TestRPCResponse2, self.broken_handler)
         self.server.register_handler("channel", "timeout", TestRPCRequest, TestRPCResponse2, self.timeout_handler)
+        self.server.add_filter(self.broken_filter, "channel", "test3")
         self.server.run_as_thread()
         self.wait_for_rpc_server(self.server)
 
-        self.client = RPCClient(self.core.messages, channels=[b"channel"], _warn=False)
+        self.client = RPCClient(self.core.messages, channels=["channel"], _warn=False)
         self.wait_for_dispatcher(self.core.messages)
 
-    def test_blocking(self):
-        result = self.client.rpc(b"channel", "test", EmptyModel(), timeout=1)
-        self.assertEqual("ok", result.status)
+    def tearDown(self):
+        self.server.shutdown()
+        super().tearDown()
 
-        with self.assertRaises(TimeoutError):
-            self.client.rpc(b"channel", "timeout", EmptyModel(), timeout=1)
+    def test_blocking(self):
+        result = self.client.rpc("channel", "test", EmptyModel(), TestRPCResponse2, timeout=1)
+        self.assertEqual("abc", result.some_payload)
+
+        with self.assertRaises(RPCClient.TimeoutError):
+            self.client.rpc("channel", "timeout", EmptyModel(), TestRPCResponse2, timeout=1)
 
     def test_non_blocking(self):
-        result = self.client.rpc(b"channel", "test", EmptyModel(), timeout=1, blocking=False)
+        result = self.client.rpc_non_blocking("channel", "test", EmptyModel(), timeout=1)
         self.assertFalse(None, result.done())
         self.assertEqual(None, result.response)
 
         time.sleep(1)
 
         self.assertTrue(result.done())
-        result = result.result()
-        self.assertEqual("ok", result.status)
-        self.assertIn("some_payload", result.response)
+        result = result.result(TestRPCResponse2)
+        self.assertIn("abc", result.some_payload)
 
     def test_non_blocking_via_result(self):
-        result = self.client.rpc(b"channel", "test", EmptyModel(), timeout=1, blocking=False)
-        result = result.result()
-        self.assertEqual("ok", result.status)
-        self.assertIn("some_payload", result.response)
+        result = self.client.rpc_non_blocking("channel", "test", EmptyModel(), timeout=1)
+        result = result.result(TestRPCResponse2)
+        self.assertIn("abc", result.some_payload)
 
     def test_non_blocking_timeout(self):
-        result = self.client.rpc(b"channel", "timeout", EmptyModel(), timeout=1, blocking=False)
-        with self.assertRaises(TimeoutError):
-            result.result(timeout=1)
+        result = self.client.rpc_non_blocking("channel", "timeout", EmptyModel(), timeout=1)
+        with self.assertRaises(RPCClient.TimeoutError):
+            result.result(TestRPCResponse2, timeout=1)
+
+    def test_non_blocking_cancel(self):
+        result = self.client.rpc_non_blocking("channel", "timeout", EmptyModel(), timeout=100)
+
+        def cancel(rpc_entry: RPCClient.RPCEntry):
+            time.sleep(0.5)
+            rpc_entry.cancel()
+
+        thread = threading.Thread(target=cancel, args=[result])
+        thread.start()
+        with self.assertRaises(RPCClient.CancelError):
+            result.result(TestRPCResponse2)
 
     def test_non_blocking_higher_timeout_than_at_request(self):
-        result = self.client.rpc(b"channel", "timeout", EmptyModel(), timeout=1, blocking=False)
-        with self.assertRaises(TimeoutError):
+        result = self.client.rpc_non_blocking("channel", "timeout", EmptyModel(), timeout=1)
+        with self.assertRaises(RPCClient.TimeoutError):
             with self.assertWarns(UserWarning) as cm:
-                result.result(timeout=2)
+                result.result(TestRPCResponse2, timeout=2)
         self.assertIn("cannot extend", str(cm.warnings[0].message))
 
     def test_non_blocking_no_timeout(self):
-        result = self.client.rpc(b"channel", "timeout", EmptyModel(), timeout=4, blocking=False)
-        result = result.result()
-        self.assertEqual("ok", result.status)
-        self.assertIn("some_payload", result.response)
+        result = self.client.rpc_non_blocking("channel", "timeout", EmptyModel(), timeout=4)
+        result = result.result(TestRPCResponse2)
+        self.assertIn("abc", result.some_payload)
 
     def test_proactive_timeout(self):
         """
         tests a case where the rpc result is cehcked after the timeout, in this case
         we time-out immediately
         """
-        result = self.client.rpc(b"channel", "timeout", EmptyModel(), timeout=1, blocking=False)
+        result = self.client.rpc_non_blocking("channel", "timeout", EmptyModel(), timeout=1)
         time.sleep(2)
         start = time.time()
-        with self.assertRaises(TimeoutError):
-            result.result()
+        with self.assertRaises(RPCClient.TimeoutError):
+            result.result(TestRPCResponse2)
         duration = time.time() - start
         self.assertTrue(duration < 0.5)
 
@@ -573,11 +606,11 @@ class RpcClientTest(RPCServerTestMixin, OpenModuleCoreTestMixin):
         because the cleanup will clean the message internally anways, there is no point in
         waiting for a longer duration
         """
-        result = self.client.rpc(b"channel", "does-not-exist", EmptyModel(), timeout=4, blocking=False)
+        result = self.client.rpc_non_blocking("channel", "does-not-exist", EmptyModel(), timeout=4)
         time.sleep(2)
         start = time.time()
-        with self.assertRaises(TimeoutError):
-            result.result()
+        with self.assertRaises(RPCClient.TimeoutError):
+            result.result(TestRPCResponse2)
         duration = time.time() - start
         # we waited 2 seconds before starting the result() call
         # thus the result call can only wait ~2 seconds, since the total timeout is 4 seconds
@@ -589,12 +622,66 @@ class RpcClientTest(RPCServerTestMixin, OpenModuleCoreTestMixin):
         but the response was received _within_ the timeout, that the
         result is still retrievable
         """
-        result = self.client.rpc(b"channel", "test", EmptyModel(), timeout=1, blocking=False)
+        result = self.client.rpc_non_blocking("channel", "test", EmptyModel(), timeout=1)
         time.sleep(3)
         self.assertTrue(result.done())
-        result = result.result()
-        self.assertEqual("ok", result.status)
-        self.assertIn("some_payload", result.response)
+        result = result.result(TestRPCResponse2)
+        self.assertIn("abc", result.some_payload)
+
+    def test_client_parse_error(self):
+        with self.assertRaises(RPCClient.ValidationError):
+            self.client.rpc("channel", "test", EmptyModel(), RPCRequest)
+
+    def test_server_parse_error(self):
+        with self.assertRaises(RPCClient.ServerValidationError):
+            self.client.rpc("channel", "test2", EmptyModel(), RPCRequest)
+
+    def test_server_filter_error(self):
+        with self.assertRaises(RPCClient.ServerFilterError):
+            self.client.rpc("channel", "test3", EmptyModel(), RPCRequest)
+
+    def test_server_handler_error(self):
+        with self.assertRaises(RPCClient.ServerHandlerError):
+            self.client.rpc("channel", "test4", EmptyModel(), RPCRequest)
+
+    def test_rpc_server_error(self):
+        with self.assertRaises(RPCClient.RPCServerError) as e:
+            entry = RPCClient.RPCEntry(100)
+            entry.response = {"status": "error", "error": "some_error"}
+            entry.result(TestRPCResponse2)
+        self.assertEqual(str(e.exception), "some_error")
+        self.assertIsInstance(e.exception, RPCClient.Exception)
+
+    def test_common_exception_type(self):
+        try:
+            self.client.rpc("channel", "test", EmptyModel(), RPCRequest)
+        except RPCClient.Exception:
+            pass
+        try:
+            self.client.rpc("channel", "test2", EmptyModel(), RPCRequest)
+        except RPCClient.Exception:
+            pass
+        try:
+            self.client.rpc("channel", "test3", EmptyModel(), RPCRequest)
+        except RPCClient.Exception:
+            pass
+        try:
+            self.client.rpc("channel", "test4", EmptyModel(), RPCRequest)
+        except RPCClient.Exception:
+            pass
+        try:
+            self.client.rpc("channel", "timeout", EmptyModel(), TestRPCResponse2, timeout=1)
+        except RPCClient.Exception:
+            pass
+
+    def test_error_and_empty_response(self):
+        with self.assertRaises(RPCClient.ServerHandlerError):
+            self.client.rpc("channel", "test4", EmptyModel(), OpenModuleModel)
+
+    def test_no_status(self):
+        rpc_entry = RPCClient.RPCEntry(1)
+        rpc_entry.response = {}
+        rpc_entry.result(OpenModuleModel)
 
 
 class TestRPCResponseNotSerializable(OpenModuleModel):
@@ -636,8 +723,8 @@ class RPCServerMultithreadingTest(OpenModuleCoreTestMixin, TestCase):
         ensures that when using a multiprocessing executor exceptions are logged to logging and sentry
         """
         with self.assertLogs(level=logging.ERROR) as cm:
-            self.rpc("test", "test", EmptyModel(), receive_response=False)
-            self.zmq_client.wait_for_message_on_topic(b"sentry")
+            self.rpc_no_response("test", "test", EmptyModel())
+            self.zmq_client.wait_for_message_on_topic("sentry")
         self.assertTrue(cm.output)
 
 
